import asyncio
import hmac
import itertools
import logging
import os
import struct
from functools import partial
from typing import Any, Callable, Optional, Tuple

from aiokdb import KException, KObj, MessageType, TypeEnum, b9, d9, krr, logger


class CredentialsException(Exception):
    pass


# TypeAlias for Optional KObj callback
OptKcb = Optional[Callable[[KObj], None]]


class KdbReader:
    def __init__(self, reader: asyncio.StreamReader, raise_krr: bool = True):
        self.reader = reader
        self.raise_krr = raise_krr

    async def read(self) -> Tuple[MessageType, KObj]:
        msgh = await self.reader.readexactly(8)
        ver, msgtype, flags, msglen = struct.unpack("<BBHI", msgh)
        logger.debug(
            f"> recv ver={ver} msgtype={msgtype} flags={flags} msglen={msglen}"
        )
        payload = await self.reader.readexactly(msglen - 8)
        k = d9(msgh + payload)
        if self.raise_krr and k.t == TypeEnum.KRR:
            raise KException(k.aS())
        return msgtype, k


class KdbWriter:
    def __init__(
        self,
        writer: asyncio.StreamWriter,
        kreader: KdbReader,
        version: int = 0,
        qid: Any = None,
    ):
        self.writer = writer
        self.qid = qid
        self.version = version
        self.reader = kreader

    def write(self, obj: KObj, mt: MessageType = MessageType.SYNC) -> None:
        bs = b9(obj, msgtype=mt)
        logger.debug(f"< sending {bs!r}")
        self.writer.write(bs)

    async def sync_req(self, obj: KObj, ooob: OptKcb = None) -> KObj:
        # responses arrive in the order that requests are sent
        # if tasks are calling this method along with read() directly
        # then responses will get muddled up or lost
        #
        # ooob is out-of-order-buffer, to optionally capture any async messages
        # or sync requests found while waiting for our response, e.g.
        #
        #  buffer = collections.deqeue(maxlen=10)
        #  await wr.sync_req(obj, ooob=buffer.append)
        #
        self.write(obj)
        while True:
            msgtype, k = await self.reader.read()
            if msgtype == MessageType.RESPONSE:
                return k
            # store ASYNC messages, or SYNC requests from the server
            if ooob is not None:
                ooob(k)
            else:
                logging.warning(f"recieved {msgtype} whilst awaiting response.")

    async def async_msg(self, obj: KObj) -> None:
        # this method is a shortcut to avoid having to import MessageType
        self.write(obj, MessageType.ASYNC)

    def close(self) -> None:
        self.writer.close()

    async def wait_closed(self) -> None:
        return await self.writer.wait_closed()


class ServerContext:
    # TODO: wrap clients in a class for timers/callbacks?

    def __init__(self, password: Optional[str] = None):
        self.password = password

    def check_login(self, user: str, password: Optional[str]) -> bool:  # .z.pw
        if self.password is None:
            return True
        # Resists timing attacks on comparison but still leaks password length
        if password is None:
            return False
        return hmac.compare_digest(self.password, password)

    async def on_sync_request(self, cmd: KObj, dotzw: KdbWriter) -> KObj:  # .z.pg
        # kdb clients usually present RPC to server as a string, evaluated
        # with value, although it is possible to have arbitary objects here
        raise Exception("nyi handling")

    async def on_async_message(self, cmd: KObj, dotzw: KdbWriter) -> None:  # .z.ps
        pass

    async def sync_response(self, cmd: KObj, dotzw: KdbWriter) -> None:
        pass

    def writer_available(self, dotzw: KdbWriter) -> None:
        # .z.po
        pass

    def writer_closing(self, dotzw: KdbWriter) -> None:
        # .z.pc
        pass

    async def start_tasks(self) -> None:
        pass


connection_counter = itertools.count()


async def process_login(
    qid: str,
    context: ServerContext,
    reader: asyncio.StreamReader,
    writer: asyncio.StreamWriter,
) -> Tuple[KdbReader, KdbWriter]:
    data = await reader.readuntil(separator=b"\000")
    # format is user(:password){ver}\000  where ver is an optional single byte
    # values currently in use are 1,2,3 ie. all non-printable ascii.
    # very old client/servers did not send a protocol version
    if len(data) < 2:
        raise Exception(f"{qid} bad login - insufficent bytes before null")
    if data[-2] < 32:
        ver = data[-2]
        auth = data[0:-2].decode()
    else:
        ver = 0
        auth = data[0:-1].decode()
    password: Optional[str] = None
    if ":" in auth:
        user, password = auth.split(":", maxsplit=1)
    else:
        user = auth
    pwstars = "*" * len(password) if password is not None else password
    logging.info(f"{qid} process_login ver={ver} user={user} password={pwstars}")

    if not context.check_login(user, password):
        raise CredentialsException("login check failed")
    writer.write(b"\x03")
    await writer.drain()

    q_reader = KdbReader(reader)
    q_writer = KdbWriter(writer, q_reader, version=ver, qid=qid)
    return q_reader, q_writer


async def handle_connection(
    context: ServerContext, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
) -> None:
    qid: str = f"q-{next(connection_counter)}"
    disconnect_log_level = logging.DEBUG
    try:
        logging.debug(f"{qid} new connection")
        q_reader, q_writer = await asyncio.wait_for(
            process_login(qid, context, reader, writer), timeout=10
        )
        context.writer_available(q_writer)
        disconnect_log_level = logging.INFO
        try:
            while not writer.is_closing():
                mtype, cmd = await q_reader.read()
                if mtype == MessageType.SYNC:
                    logging.info(f"{qid} command {cmd}")
                    try:
                        q_writer.write(
                            await context.on_sync_request(cmd, q_writer),
                            MessageType.RESPONSE,
                        )
                    except asyncio.TimeoutError as e:
                        logging.info("Sync command had timeout, continue")
                        q_writer.write(krr(str(e)), MessageType.RESPONSE)
                    except Exception as e:
                        logging.warning(
                            f"sync command {cmd} resulted in exception {repr(e)}",
                            exc_info=True,
                        )
                        q_writer.write(krr(str(e)), MessageType.RESPONSE)

                elif mtype == MessageType.ASYNC:
                    try:
                        await context.on_async_message(cmd, q_writer)
                    except asyncio.TimeoutError:
                        logging.info("Async command had timeout, continue")
                    except Exception as e:
                        logging.warning(
                            f"async command {cmd} resulted in exception {repr(e)}",
                            exc_info=True,
                        )
                elif mtype == MessageType.RESPONSE:
                    try:
                        await context.sync_response(cmd, q_writer)
                    except Exception as e:
                        logging.warning(
                            f"response processing resulted in exception: {repr(e)}",
                            exc_info=True,
                        )
                else:
                    raise Exception(f"{qid} Unexpected incoming message type")
        finally:
            context.writer_closing(q_writer)

    except asyncio.TimeoutError:
        logging.info(f"{qid} closed - login timeout")
    except (
        asyncio.exceptions.IncompleteReadError
    ):  # When kdb process timeout via .timer.timeoutSyncCall
        logging.log(disconnect_log_level, f"{qid} connection reached end of stream")
    except BrokenPipeError:
        logging.info(f"{qid} connection ended BrokenPipeError")
    except CredentialsException:
        logging.info(f"{qid} login credentials incorrect, closed")
    finally:
        try:
            # throws RuntimeError in test teardown
            writer.close()
        except RuntimeError:
            pass


async def start_qserver(
    port: int, context: ServerContext, periodic: bool = False
) -> Any:
    logging.info(f"opening KDB-q IPC server on port {port}")
    if periodic:
        await context.start_tasks()

    return await asyncio.start_server(partial(handle_connection, context), "", port)


async def main(qpassword: str, qport: int) -> None:
    context = ServerContext(qpassword)
    server = await start_qserver(qport, context)
    async with server:
        await server.serve_forever()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import argparse

    parser = argparse.ArgumentParser()
    QPORT_DEFAULT = os.environ.get("QPORT", 8890)
    QPASSWORD_DEFAULT = os.environ.get("QPASSWORD", None)
    parser.add_argument(
        "--qport",
        default=QPORT_DEFAULT,
        help="port to listen for kdb-q IPC connections",
    )
    parser.add_argument(
        "--qpassword",
        default=QPASSWORD_DEFAULT,
        help="password for kdb-q IPC connections",
    )
    args = parser.parse_args()

    asyncio.run(main(args.qpassword, args.qport))



import argparse
import asyncio
import logging
import os
import traceback
from typing import Any

from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.shortcuts import PromptSession

from aiokdb import TypeEnum, cv
from aiokdb.client import open_qipc_connection
from aiokdb.format import AsciiFormatter


async def main(args: Any) -> None:
    r, w = await open_qipc_connection(
        host=args.host, port=args.port, user=args.user, password=args.password
    )

    history = FileHistory(os.path.expanduser("~/.aiokdb-cli-history"))
    session: Any = PromptSession("(eval) > ", history=history)
    fmt = AsciiFormatter(height=args.height)

    # Run echo loop. Read text from stdin, and reply it back.
    while True:
        try:
            inp = await session.prompt_async()
            if inp == "":
                continue
            output = await w.sync_req(cv(inp))
            if output.t != TypeEnum.NIL:
                print(fmt.format(output))
        except KeyboardInterrupt:
            return
        except EOFError:
            break
        except Exception:
            traceback.print_exc(limit=-2)
    return None


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="localhost")
    parser.add_argument("--port", default=8890, type=int)
    parser.add_argument("--user", default="user")
    parser.add_argument("--password")
    parser.add_argument("--height", default=10, type=int)
    parser.add_argument("--debug", action="store_true")

    args = parser.parse_args()
    debug_level = {True: logging.DEBUG, False: logging.INFO}
    logging.basicConfig(level=debug_level[args.debug])
    with patch_stdout():
        asyncio.run(main(args))



from typing import List

import pytest

from aiokdb import KException, KObj, MessageType, cv, kj
from aiokdb.client import open_qipc_connection
from aiokdb.server import CredentialsException, KdbWriter, ServerContext, start_qserver


@pytest.mark.asyncio
async def test_client_noauth_defaultport() -> None:
    context = ServerContext()
    server = await start_qserver(8890, context)

    client_rd, client_wr = await open_qipc_connection()

    with pytest.raises(KException, match="nyi handling"):
        await client_wr.sync_req(cv("1+2"))

    client_wr.close()
    await client_wr.wait_closed()

    # check supplied auth ignored when noauth configured
    client_rd, client_wr = await open_qipc_connection(user="troy", password="tango")
    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()


# Please manually verify auth changes against KDB.
#
# Python server, KDB client
#   $ python -m qaio.server
#   $ ~/kdb/q
#   q)hopen ":localhost:8890"           INFO:root:q-0 process_login ver=6 user=chrisshucksmith password=None
#   q)hopen ":localhost:8890:blah"      INFO:root:q-1 process_login ver=6 user=blah password=None
#   q)hopen ":localhost:8890:blah:boo"  INFO:root:q-2 process_login ver=6 user=blah password=***
#
# Python client, KDB server
#   $ ~/kdb/q -p 8890
#   q).z.pw:{[user;pswd]1 "user: ",(string user)," password: ",pswd,"\n";1b}
#   $ python -m qaio.client                          q) user: user password:
#   $ python -m qaio.client --user u1                q) user: u1 password:
#   $ python -m qaio.client --user u1 --password p2  q) user: u1 password: p2


@pytest.mark.asyncio
async def test_client_auth_defined_port() -> None:
    # echo back client request
    class EchoServerContext(ServerContext):
        async def on_sync_request(self, cmd: KObj, dotzw: KdbWriter) -> KObj:
            return cmd

    context = EchoServerContext("tangoxray")
    server = await start_qserver(6778, context)

    # check correct auth succeeds
    client_rd, client_wr = await open_qipc_connection(
        port=6778, user="troy", password="tangoxray"
    )
    kr = await client_wr.sync_req(cv("1+2"))
    assert kr.aS() == "1+2"

    client_wr.close()
    await client_wr.wait_closed()

    # check wrong auth fails
    with pytest.raises(CredentialsException):
        client_rd, client_wr = await open_qipc_connection(
            port=6778, user="troy", password="xyz"
        )

    # hopen can be sent without a password
    with pytest.raises(CredentialsException):
        client_rd, client_wr = await open_qipc_connection(
            port=6778, user="xyz", password=None
        )

    # hopen can be sent with empty username (encodes as "")
    with pytest.raises(CredentialsException):
        client_rd, client_wr = await open_qipc_connection(
            port=6778, user=None, password=None
        )

    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_server_async() -> None:
    class AsyncServerContext(ServerContext):
        async def on_sync_request(self, cmd: KObj, dotzw: KdbWriter) -> KObj:
            # async message sent during sync request handling
            dotzw.write(kj(1), MessageType.ASYNC)
            # return value becomes the sync response
            return kj(2)

    context = AsyncServerContext("tangoxray")
    server = await start_qserver(6778, context)

    client_rd, client_wr = await open_qipc_connection(
        port=6778, user="troy", password="tangoxray"
    )

    ooob: List[KObj] = []
    kr = await client_wr.sync_req(cv("1+2"), ooob=ooob.append)
    assert kr.aJ() == 2

    assert len(ooob) == 1
    assert ooob[0].aJ() == 1

    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()



import logging
import socket
import struct
from typing import Optional

from aiokdb import KException, KObj, MessageType, TypeEnum, b9, cv, d9


# I don't recommend using, the asyncio interface is way nicer
class KSocket:
    def __init__(self, skt: socket.socket, raise_krr: bool = True):
        self.s = skt
        self.raise_krr = raise_krr

    def k(self, cmd: str, data: Optional[KObj] = None) -> KObj:
        ko = cv(cmd)
        self.s.sendall(b9(ko, msgtype=MessageType.SYNC))
        msgh = self.readexactly(8)
        ver, msgtype, flags, msglen = struct.unpack("<BBHI", msgh)
        print(f"> recv ver={ver} msgtype={msgtype} flags={flags} msglen={msglen}")
        payload = self.readexactly(msglen - 8)

        k = d9(msgh + payload)
        if self.raise_krr and k.t == TypeEnum.KRR:
            raise KException(k.aS())
        return k

    def readexactly(self, sz: int) -> bytes:
        bs = b""
        while len(bs) < sz:
            bs = bs + self.s.recv(sz - len(bs))
        return bs


def khpu(
    host: str = "localhost",
    port: int = 12345,
    auth: str = "kdb:pass",
    ver: int = 3,
    raise_krr: bool = True,
) -> KSocket:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    s.sendall(auth.encode() + struct.pack("<B", ver) + b"\000")
    data = s.recv(1)
    remote_ver = struct.unpack("<B", data)[0]
    if remote_ver != ver:
        raise Exception(f"expected version {ver}, server gave {remote_ver}")
    return KSocket(s, raise_krr=raise_krr)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    h = khpu("localhost", 12345, "kdb:pass")
    result = h.k("2.0+3.0")
    assert result.aF() == 5.0
    print(result)
    result.aJ()



import os
from urllib.parse import urlparse

import pytest

from aiokdb import Infs, KObj, Nulls, TypeEnum, cv
from aiokdb.client import open_qipc_connection
from aiokdb.format import AsciiFormatter

# test requires working KDB's eval in the server process, so can't use our python shim


@pytest.mark.asyncio
async def test_rpc() -> None:
    server_uri = os.environ.get("KDB_PYTEST_SERVICE", None)

    if not server_uri:
        pytest.skip("No KDB_PYTEST_SERVICE provided")
        return

    o = urlparse(server_uri)
    h = o.hostname
    p = o.port
    if h is None or p is None:
        raise Exception("url must provide kdb://hostname:port")
    krd, kwr = await open_qipc_connection(
        host=h, port=p, user=o.username, password=o.password
    )

    fmt = AsciiFormatter()

    async def check_result(cmd: str, t: int) -> KObj:
        k = await kwr.sync_req(cv(cmd))
        assert k.t == t
        fmt.format(k)
        return k

    # atoms
    await check_result("0b", -TypeEnum.KB)
    await check_result("0Ng", -TypeEnum.UU)
    await check_result("0x00", -TypeEnum.KG)
    await check_result("0h", -TypeEnum.KH)
    await check_result("0i", -TypeEnum.KI)
    await check_result("0j", -TypeEnum.KJ)
    await check_result("0e", -TypeEnum.KE)
    await check_result("0.0", -TypeEnum.KF)
    await check_result('"x"', -TypeEnum.KC)
    await check_result("`", -TypeEnum.KS)
    await check_result("2000.01.01D00:00:00.000000000", -TypeEnum.KP)
    await check_result("2000.01m", -TypeEnum.KM)
    await check_result("1984.01.25", -TypeEnum.KD)
    # deprecated floating point
    await check_result("2000.01.01T00:00:00.000", -TypeEnum.KZ)
    await check_result("00:00:00.000000000", -TypeEnum.KN)
    await check_result("23:59", -TypeEnum.KU)
    await check_result("23:59:00", -TypeEnum.KV)
    await check_result("23:59:00.000", -TypeEnum.KT)

    # vectors
    await check_result("enlist 0b", TypeEnum.KB)
    await check_result("enlist 0Ng", TypeEnum.UU)
    await check_result("enlist 0x00", TypeEnum.KG)
    await check_result("enlist 0h", TypeEnum.KH)
    await check_result("enlist 0i", TypeEnum.KI)
    await check_result("enlist 0j", TypeEnum.KJ)
    await check_result("enlist 0e", TypeEnum.KE)
    await check_result("enlist 0.0", TypeEnum.KF)
    await check_result('enlist "x"', TypeEnum.KC)
    await check_result("enlist `", TypeEnum.KS)
    await check_result("enlist 2000.01.01D00:00:00.000", TypeEnum.KP)
    await check_result("enlist 2000.01m", TypeEnum.KM)
    await check_result("enlist 1984.01.25", TypeEnum.KD)
    await check_result("enlist 2000.01.01T00:00:00.000", TypeEnum.KZ)
    await check_result("enlist 00:00:00.000000000", TypeEnum.KN)
    await check_result("enlist 23:59", TypeEnum.KU)
    await check_result("enlist 23:59:00", TypeEnum.KV)
    await check_result("enlist 23:59:00.000", TypeEnum.KT)

    await check_result("`a`b`c!(1 2;3 5;7 11)", TypeEnum.XD)
    await check_result("flip `a`b`c!(1 2;3 5;7 11)", TypeEnum.XT)
    await check_result("`s#4 5 6!6 5 4", TypeEnum.SD)

    # fancy
    await check_result("::", TypeEnum.NIL)
    await check_result("", TypeEnum.NIL)

    # await check_result(('{x+y}', TypeEnum.FN
    # https://code.kx.com/q/basics/datatypes/#functions-iterators-derived-functions
    # q)type each({x+y};neg;-;\;+[;1];<>;,';+/;+\;prev;+/:;+\:;`f 2:`f,1)
    # 100 101 102 103 104 105 106 107 108 109 110 111 112h

    # nulls
    assert (await check_result("0Nh", -TypeEnum.KH)).aH() == Nulls.h
    # await check_result(("0Ng"))).aU() == ku(Nulls.u)
    assert (await check_result("0Ni", -TypeEnum.KI)).aI() == Nulls.i
    assert (await check_result("0Nj", -TypeEnum.KJ)).aJ() == Nulls.j
    # await check_result(("0Ne"))).aE() == Nulls.e
    # await check_result(("0Nf"))).aF() == Nulls.f

    # infs
    assert (await check_result("0Wh", -TypeEnum.KH)).aH() == Infs.h
    assert (await check_result("0Wi", -TypeEnum.KI)).aI() == Infs.i
    assert (await check_result("0Wj", -TypeEnum.KJ)).aJ() == Infs.j
    assert (await check_result("0We", -TypeEnum.KE)).aE() == Infs.e
    assert (await check_result("0w", -TypeEnum.KF)).aF() == Infs.f

    await check_result(
        's:(0b;0Ng;0x00;0h;0i;0j;0e;0.0;"x";`;2000.01.01D;2023.01m;2023.12.23;2000.01.01T;00:00:00.000000000;23:59;23:59:00;23:59:00.000);([]t:type each s;atom:s;vect:enlist each s)',
        TypeEnum.XT,
    )



import itertools
from datetime import datetime, timezone
from typing import Iterable, List, Optional, Sequence

from aiokdb import KObj, Nulls, TypeEnum

# Keeping formating external to KObj since it complicates the implementation.
# This produces AsciiFormatter produces numpy-like formatting, ie. it truncates the central section of a table
#
#    a b| addr e  f
#    ---|----------
#    2 1| 1    44 7
#    ...|...
#    3 4| abc  34 4
#
# allows customisation by subclassing
# TODO: truncate width


class AsciiFormatter:
    def __init__(self, width: int = 200, height: int = 10):
        self.width = width
        self.height = height
        self.inline_chars = width / 3

    def format(self, obj: KObj) -> str:
        if obj.t == TypeEnum.XT:
            return self._fmt_unkeyed_table(obj)
        elif obj.t == TypeEnum.XD and obj.kkey().t == TypeEnum.XT:
            return self._fmt_keyed_table(obj)
        elif obj.t == TypeEnum.XD:
            return self._fmt_dict(obj)
        return self._fmt_inline(obj)

    def _fmt_unkeyed_table(self, obj: KObj) -> str:
        rowcount = self._table_conforms(obj)
        rows = self._select_rows(rowcount)
        return "\n".join(self._format_table_rows(obj, rows))

    def _select_rows(self, rowcount: int) -> Iterable[Optional[int]]:
        if rowcount < self.height - 2:
            # no need to truncate rows
            rows: Iterable[Optional[int]] = itertools.chain(range(rowcount))
        else:
            chunk = (self.height - 3) // 2
            rows = itertools.chain(
                range(chunk), [None], range(rowcount - chunk, rowcount)
            )
        return rows

    def _table_conforms(self, obj: KObj) -> int:
        assert obj.t == TypeEnum.XT
        obj = obj.kvalue()
        assert obj.t == TypeEnum.XD
        assert obj.kkey().t == TypeEnum.KS
        assert obj.kvalue().t == TypeEnum.K

        # check rows conform
        kv = obj.kvalue().kK()
        rowcount = len(kv[0])
        for i in range(len(obj.kkey())):
            assert len(kv[i]) == rowcount
        return rowcount

    def _fmt_keyed_table(self, obj: KObj) -> str:
        ktv = obj.kvalue()
        ktk = obj.kkey()

        # check rows conform
        keyrowcount = self._table_conforms(ktv)
        valuerowcount = self._table_conforms(ktk)

        assert keyrowcount == valuerowcount
        rows = list(self._select_rows(keyrowcount))

        left = self._format_table_rows(obj.kkey(), rows)
        right = self._format_table_rows(obj.kvalue(), rows)

        filler = [" ", "-"] + [" "] * len(rows)
        return "\n".join(f"{ll}|{g}{rr}" for g, ll, rr in zip(filler, left, right))

    def _format_table_rows(self, obj: KObj, rows: Iterable[Optional[int]]) -> List[str]:
        d1 = obj.kvalue()

        colNames: Sequence[str] = d1.kkey().kS()
        kv = d1.kvalue().kK()

        colWidths: List[int] = list(map(len, colNames))

        # stringify all cells within our rowiter
        rowSample: List[List[str]] = []
        for r in rows:
            cs = []
            for c in range(len(colNames)):
                s = self._str_cell(kv[c], c, r)
                colWidths[c] = max(colWidths[c], len(s))
                cs.append(s)
            rowSample.append(cs)

        # now do the padding of pre-stringified cells
        rowText = []
        for row in rowSample:
            rowText.append(" ".join([f"{t:{w}}" for w, t in zip(colWidths, row)]))

        headers = " ".join([f"{t:{w}}" for w, t in zip(colWidths, colNames)])
        dashes = "".join(["-"] * len(headers))
        return [headers, dashes, *rowText]

    def _str_cell(self, obj: KObj, col: int, index: Optional[int]) -> str:
        if index is None:
            if col == 0:
                return "..."
            return ""

        if obj.t == TypeEnum.KG or obj.t == TypeEnum.KB:
            return str(obj.kG()[index])
        elif obj.t == TypeEnum.KH:
            i = obj.kH()[index]
            if i == Nulls.h:
                return ""
            return str(i)
        elif obj.t == TypeEnum.KI:
            i = obj.kI()[index]
            if i == Nulls.i:
                return ""
            return str(i)
        elif obj.t == TypeEnum.KJ:
            j = obj.kJ()[index]
            if j == Nulls.j:
                return ""
            if j == 9223372036854775807:
                return "0W"
            return str(j)
        elif obj.t == TypeEnum.KM:
            return self._fmt_atom_m(obj.kI()[index])
        elif obj.t == TypeEnum.KD:
            return self._fmt_atom_d(obj.kI()[index])
        elif obj.t == TypeEnum.KZ:
            return self._fmt_atom_z(obj.kF()[index])
        elif obj.t == TypeEnum.KU:
            return self._fmt_atom_u(obj.kI()[index])
        elif obj.t == TypeEnum.KV:
            return self._fmt_atom_v(obj.kI()[index])
        elif obj.t == TypeEnum.KT:
            return self._fmt_atom_t(obj.kI()[index])
        elif obj.t == TypeEnum.KS:
            return obj.kS()[index]
        elif obj.t == TypeEnum.KN:
            j = obj.kJ()[index]
            return self._fmt_atom_n(j)
        elif obj.t == TypeEnum.K:
            # how much of obj should we show?
            o = obj.kK()[index]
            return self._fmt_inline(o)
        elif obj.t == TypeEnum.UU:
            return str(obj.kU()[index])
        elif obj.t == TypeEnum.KE:
            e = obj.kE()[index]
            return str(e)
        elif obj.t == TypeEnum.KF:
            return str(obj.kF()[index])
        elif obj.t == TypeEnum.KP:
            j = obj.kJ()[index]
            return self._fmt_atom_p(j)
        elif obj.t == TypeEnum.KC:
            return obj.aS()[index]
        raise ValueError(f"No cell formatter for {obj} with type {obj._tn()}")

    def _fmt_atom_p(self, j: int) -> str:
        if j == Nulls.j:
            return ""
        # timestamp (nanos) q)"p"$1  2000.01.01D00:00:00.000000001
        # python timestamps have microsecond precision, so on our own
        # formatting with full precision
        nanos = j % 1000
        micros = j // 1000
        origin = int(datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp())
        dt = datetime.utcfromtimestamp(origin + micros / 1000000.0)
        return dt.strftime("%Y.%m.%dD%H:%M:%S:%f") + f"{nanos:03}"

    def _fmt_atom_n(self, j: int) -> str:
        if j == Nulls.j:
            return ""
        # timespan (nanos) q) "n"$1  0D00:00:00.000000001
        secs = j // 1000000000
        m = secs // 60
        h = m // 60
        d = h // 24
        nanos = j % 1000000000
        secs = secs % 60
        m = m % 60
        h = h % 24
        return f"{d}D{h:02}:{m:02}:{secs:02}.{nanos:09}"

    def _fmt_atom_m(self, m: int) -> str:
        if m == Nulls.i:
            return "0Nm"
        return f"{m/12:04}.{m%12:02}m"

    def _fmt_atom_d(self, d: int) -> str:
        if d == Nulls.i:
            return "0Nd"
        origin = int(datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp())
        dt = datetime.utcfromtimestamp(origin + d)
        return dt.strftime("%Y.%m.%d")

    def _fmt_atom_z(self, d: float) -> str:
        if d == Nulls.f:
            return "0Nz"
        origin = int(datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp())
        dt = datetime.utcfromtimestamp(origin + d)
        return dt.strftime("%Y.%m.%dT%H:%M:%S:%f")

    def _fmt_atom_u(self, u: int) -> str:
        if u == Nulls.i:
            return "0Nu"
        # TODO: formatting >1h etc
        return f"{u/60:02}:{u%60:02}"

    def _fmt_atom_v(self, v: int) -> str:
        if v == Nulls.i:
            return "0Nv"
        h = v / 3600
        m = (v / 60) % 60
        s = v % 60
        return f"{h:02}:{m:02}:{s:02}"

    def _fmt_atom_t(self, t: int) -> str:
        if t == Nulls.i:
            return "ONt"
        # timespan in nanos
        secs = t // 1000000000
        m = secs // 60
        h = m // 60
        d = h // 24
        nanos = t % 1000000000
        secs = secs % 60
        m = m % 60
        h = h % 24
        return f"{d}D{h:02}:{m:02}:{secs:02}.{nanos:09}"

    def _fmt_dict(self, obj: KObj) -> str:
        # measure the keys
        rows = list(self._select_rows(len(obj)))
        ks, vs = [], []
        keywidth = 0
        for r in rows:
            k = self._str_cell(obj.kkey(), 0, r)
            v = self._str_cell(obj.kvalue(), 0, r)
            keywidth = max(keywidth, len(k))
            ks.append(k)
            vs.append(v)

        return "\n".join([f"{k:{keywidth}}| {v}" for k, v in zip(ks, vs)])

    def _fmt_inline(self, obj: KObj) -> str:
        # we are expecting to format the object in a constrained environment, ie. within the
        # cell of a table or dictionary. We can take up at most one line of text and width
        # of inline_chars
        if obj.t == -TypeEnum.KJ:
            j = obj.aJ()
            if j == Nulls.j:
                return ""
            if j == 9223372036854775807:
                return "0W"
            return str(j)
        elif obj.t == -TypeEnum.KI:
            i = obj.aI()
            if i == Nulls.i:
                return ""
            return str(i) + "i"
        elif obj.t == -TypeEnum.KH:
            i = obj.aH()
            return str(i) + "h"
        elif obj.t == -TypeEnum.KE:
            e = obj.aE()
            return str(e) + "e"
        elif obj.t == -TypeEnum.KS:
            return obj.aS()
        elif obj.t == -TypeEnum.KC:
            return f'"{obj.aC()}"'
        elif obj.t == -TypeEnum.KG or obj.t == -TypeEnum.KB:
            return str(obj.aG())
        elif obj.t == -TypeEnum.KN:
            return self._fmt_atom_n(obj.aJ())
        elif obj.t == -TypeEnum.KM:
            return self._fmt_atom_m(obj.aI())
        elif obj.t == -TypeEnum.KD:
            return self._fmt_atom_d(obj.aI())
        elif obj.t == -TypeEnum.KZ:
            return self._fmt_atom_z(obj.aF())
        elif obj.t == -TypeEnum.KU:
            return self._fmt_atom_u(obj.aI())
        elif obj.t == -TypeEnum.KV:
            return self._fmt_atom_v(obj.aI())
        elif obj.t == -TypeEnum.KT:
            return self._fmt_atom_t(obj.aI())
        elif obj.t == -TypeEnum.UU:
            return str(obj.aU())
        elif obj.t == -TypeEnum.KF:
            return str(obj.aF())
        elif obj.t == -TypeEnum.KP:
            return self._fmt_atom_p(obj.aJ())

        # vectors
        elif obj.t == TypeEnum.K:
            # sample the vector (first five?)
            elems = list(self._select_rows(len(obj)))
            ks = ", ".join([self._str_cell(obj, 0, r) for r in elems])
            return f"[{ks}]"
        elif obj.t == TypeEnum.KC:
            elems = list(self._select_rows(len(obj)))
            return "".join([self._str_cell(obj, 0, r) for r in elems])

        elif obj.t > 0 and obj.t < 20:
            # sample the vector (first five?)
            elems = list(self._select_rows(len(obj)))
            ks = " ".join([self._str_cell(obj, 0, r) for r in elems])
            return ks

        elif obj.t in (TypeEnum.XD, TypeEnum.SD):
            return "KDict"
        elif obj.t == TypeEnum.XT:
            return "KTable"
        elif obj.t == TypeEnum.NIL:
            return "::"

        raise ValueError(f"No inline formatter for {obj} with type {obj._tn()}")
